/**
 * $Id: Field.java 33 2006-03-22 18:22:13Z Gerhard $
 * 
 * Copyright (C) 2006 Gerhard Zlabinger (gerhard.zlabinger@gmail.com)
 *
 * This file is part of Ubertris.
 * 
 * Ubertris is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ubertris is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ubertris; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

package ubertris;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

final class Field {

    /*
     * field dimensions
     */

    public static final int     WIDTH  = 10;
    public static final int     HEIGHT = 20;

    /*
     * constants
     */
    public static final int     FIELD_EMPTY    = 0;

    /*
     * field coordinates
     */

    public int                  SCREEN_TOP;
    public int                  SCREEN_HEIGHT;
    public int                  SCREEN_WIDTH;

    /*
     * preview area
     */

    public int                  PREVIEW_WIDTH  = 4;
    public int                  PREVIEW_HEIGHT = 4;
    public int                  PREVIEW_SCREEN_TOP;
    public int                  PREVIEW_SCREEN_LEFT;
    public int                  PREVIEW_SCREEN_WIDTH;
    public int                  PREVIEW_SCREEN_HEIGHT;

    /*
     * cube size
     */

    public int                  CUBE_WIDTH;
    public int                  CUBE_HEIGHT;

    /*
     * canvas reference
     */
    public final UbertrisCanvas canvas;
    
    /*
     * field definitions
     */

    private int[][]             field;
    private int[]               freeSlotsInRow;

    public Field(UbertrisCanvas canvas) {
        this.canvas = canvas;

        reset();
    }

    public Field(UbertrisCanvas canvas, DataInputStream dos) throws IOException {
        this.canvas = canvas;

        deserialize(dos);
    }

    public void calculateSizes() {
        /*
         * calculate screen sizes...
         */

        CUBE_HEIGHT = canvas.height / HEIGHT;
        CUBE_WIDTH = Math.min((int) ((canvas.width * 6) / 10) / WIDTH, CUBE_HEIGHT);

        SCREEN_WIDTH = CUBE_WIDTH * WIDTH;
        SCREEN_HEIGHT = CUBE_HEIGHT * HEIGHT;

        SCREEN_TOP = canvas.height - SCREEN_HEIGHT;

        /*
         * calculate preview area
         */

        PREVIEW_SCREEN_HEIGHT = PREVIEW_HEIGHT * CUBE_HEIGHT;
        PREVIEW_SCREEN_WIDTH = PREVIEW_WIDTH * CUBE_WIDTH;
        PREVIEW_SCREEN_TOP = canvas.height - PREVIEW_SCREEN_HEIGHT;
        PREVIEW_SCREEN_LEFT = SCREEN_WIDTH + ((canvas.width - SCREEN_WIDTH - PREVIEW_SCREEN_WIDTH) >> 1);
    }

    // TODO: performance by making field public
    public int getColorAt(int x, int y) {
        return field[y][x];
    }

    // TODO: performance by making field public
    /**
     * this must only be called on empty fields
     * 
     * @param x
     * @param y
     * @param color
     * @return true if the row was completed & removed
     */
    public boolean setColorAt(int x, int y, int color) {
        field[y][x] = color;

        // remove the row if full
        if (--freeSlotsInRow[y] == 0) {

            final int[] temp = field[y];
            System.arraycopy(field, 0, field, 1, y);
            field[0] = temp;

            // recycle array
            // TODO: eliminate loop
            for (int w = WIDTH - 1; w >= 0; --w)
                field[0][w] = FIELD_EMPTY;

            System.arraycopy(freeSlotsInRow, 0, freeSlotsInRow, 1, y);
            freeSlotsInRow[0] = WIDTH;

            return true;
        }

        return false;
    }

    // TODO: performance by making field public
    public boolean isCubeFree(int x, int y) {
        return field[y][x] == FIELD_EMPTY;
    }

    public void renderCubeAt(int x, int y, int color) {
        canvas.buffer.setColor(color);
        canvas.buffer.fillRect(x + 1, y + 1, CUBE_WIDTH - 2, CUBE_HEIGHT - 2);

        canvas.buffer.setColor(color | 0xBBBBBB);
        canvas.buffer.drawLine(x, y, x + CUBE_WIDTH - 1, y);
        canvas.buffer.drawLine(x, y + 1, x, y + CUBE_HEIGHT - 1);

        canvas.buffer.setColor(color & 0x060606);
        canvas.buffer.drawLine(x + CUBE_WIDTH - 1, y + 1, x + CUBE_WIDTH - 1, y + CUBE_HEIGHT - 1);
        canvas.buffer.drawLine(x, y + CUBE_HEIGHT - 1, x + CUBE_WIDTH - 1, y + CUBE_HEIGHT - 1);
    }

    public void renderCube(int x, int y, int color) {
        renderCubeAt(CUBE_WIDTH * x, SCREEN_TOP + CUBE_HEIGHT * y, color);
    }

    public void render() {
        canvas.buffer.setColor(UbertrisCanvas.COLOR_FIELD_BG);
        canvas.buffer.fillRect(0, SCREEN_TOP, SCREEN_WIDTH, SCREEN_HEIGHT);

        for (int v = 0, y = SCREEN_TOP; v < HEIGHT; ++v, y += CUBE_HEIGHT) {
            for (int u = 0, x = 0; u < WIDTH; ++u, x += CUBE_WIDTH) {
                if (field[v][u] != 0) {
                    renderCubeAt(x, y, field[v][u]);
                }
            }
        }
    }

    public void reset() {
        field = new int[HEIGHT][WIDTH];
        freeSlotsInRow = new int[HEIGHT];

        for (int i = HEIGHT - 1; i >= 0; --i)
            freeSlotsInRow[i] = WIDTH;
    }

    /*
     * tile preview
     */

    public void renderTilePreview(Tile tile) {
        int y = PREVIEW_SCREEN_TOP + (PREVIEW_SCREEN_HEIGHT - CUBE_HEIGHT * tile.height >> 1);
        final int startX = PREVIEW_SCREEN_LEFT + (PREVIEW_SCREEN_WIDTH - CUBE_WIDTH * tile.width >> 1);
        for (int v = 0; v < tile.height; ++v, y += CUBE_HEIGHT) {
            for (int u = 0, x = startX; u < tile.width; ++u, x += CUBE_WIDTH) {
                if (tile.layout[v][u]) {
                    renderCubeAt(x, y, tile.color);
                }
            }
        }
    }

    public void clearPreview() {
        canvas.buffer.setColor(UbertrisCanvas.COLOR_FIELD_BG);
        canvas.buffer.fillRect(PREVIEW_SCREEN_LEFT, PREVIEW_SCREEN_TOP, PREVIEW_SCREEN_WIDTH, PREVIEW_SCREEN_HEIGHT);
    }

    public void renderTile(Tile tile) {
        final int startX = tile.posX * CUBE_WIDTH;

        for (int v = tile.height - 1, y = SCREEN_TOP + tile.posY * CUBE_HEIGHT; v >= 0 && y >= SCREEN_TOP; --v, y -= CUBE_HEIGHT) {
            for (int u = 0, x = startX; u < tile.width; ++u, x += CUBE_WIDTH) {
                if (tile.layout[v][u]) {
                    renderCubeAt(x, y, tile.color);
                }
            }
        }
    }

    public void eraseTile(Tile tile) {
        canvas.buffer.setColor(UbertrisCanvas.COLOR_FIELD_BG);

        final int startX = tile.oldPosX * CUBE_WIDTH;
        for (int v = tile.oldHeight - 1, y = SCREEN_TOP + tile.oldPosY * CUBE_HEIGHT; v >= 0 && y >= SCREEN_TOP; --v, y -= CUBE_HEIGHT) {
            for (int u = 0, x = startX; u < tile.oldWidth; ++u, x += CUBE_WIDTH) {
                if (tile.oldLayout[v][u]) {
                    canvas.buffer.fillRect(x, y, CUBE_WIDTH, CUBE_HEIGHT);
                }
            }
        }
    }

    /*
     * Serialization
     */

    public void serialize(DataOutputStream dos) throws IOException {
        for (int j = 0; j < HEIGHT; ++j) {
            for (int i = 0; i < WIDTH; ++i) {
                dos.writeInt(field[j][i]);
            }
        }
    }

    private void deserialize(DataInputStream dis) throws IOException {
        field = new int[HEIGHT][WIDTH];
        freeSlotsInRow = new int[HEIGHT];

        for (int j = 0; j < HEIGHT; ++j) {
            int freeSlots = WIDTH;
            
            for (int i = 0; i < WIDTH; ++i) {
                field[j][i] = dis.readInt();
                
                if (field[j][i] != 0)
                  --freeSlots;
            }

            freeSlotsInRow[j] = freeSlots;
        }
    }
}
