/**
 * $Id: GameThread.java 33 2006-03-22 18:22:13Z Gerhard $
 * 
 * Copyright (C) 2006 Gerhard Zlabinger (gerhard.zlabinger@gmail.com)
 *
 * This file is part of Ubertris.
 * 
 * Ubertris is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ubertris is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ubertris; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

package ubertris;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Timer;
import java.util.TimerTask;

import javax.microedition.lcdui.Canvas;

final class GameThread extends Thread {

    /*
     * static stuff
     */

    static final int             TILES_PER_LEVEL     = 50;

    private static final int     STATE_RUNNING       = 0;
    private static final int     STATE_PAUSED        = 1;
    private static final int     STATE_DEAD          = 2;
    private static final int     STATE_FRESH         = 3;

    private static final int     INITIAL_LEVEL_DELAY = 600;

    /*
     * instance stuff
     */

    // TODO: get rid of this silly keyQueue
    public final Object          actionLock          = new Object();
    public int                   actionCode;

    protected Tile               tileCurrent;
    protected Tile               tileNext;

    int                          score;
    int                          level;
    int                          tileCount;

    final UbertrisCanvas canvas;
    Timer                timerGravity;
    int                  state               = STATE_FRESH;
    int                  stateRequested;
    int                  levelDelay;

    /*
     * construction time again
     */

    public GameThread(UbertrisCanvas canvas) {
        this(canvas, Tile.getRandomTile(canvas), Tile.getRandomTile(canvas), 0, INITIAL_LEVEL_DELAY, 0, 0);

        tileCurrent.startDescent();
        stateRequested = STATE_RUNNING;
    }

    public GameThread(UbertrisCanvas canvas, Tile tileCurrent, Tile tileNext, int level, int levelDelay, int score, int tileCount) {
        this.canvas = canvas;

        this.tileNext = tileNext;
        this.tileCurrent = tileCurrent;
        this.level = level;
        this.levelDelay = levelDelay;
        this.score = score;
        this.tileCount = tileCount;

        stateRequested = STATE_PAUSED;
    }

    /*
     * state management
     */

    private synchronized void requestState(int state) {
        if (stateRequested == state || this.state == STATE_DEAD)
            return;

        synchronized (this) {
            stateRequested = state;

            /*
             * wait until gameThread has acknowledged the new state
             */

            notifyAll();
            try {
                wait();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
    }

    private synchronized void setState(int newState) {
        if (newState != state) {
            switch (state) {
                case STATE_RUNNING:
                    timerGravity.cancel();
                    timerGravity = null;
                    break;
            }

            switch (newState) {
                case STATE_PAUSED:
                    canvas.paused();
                    break;

                case STATE_RUNNING:
                    canvas.completelyRedrawScreen();
                    canvas.flushGraphics();

                    timerGravity = new Timer();
                    timerGravity.scheduleAtFixedRate(new TimerTask() {

                        public void run() {
                            commitDescent();
                        }

                    }, 0, levelDelay);
                    break;
            }

            state = newState;
        }
        /*
         * notify requestState() of the state change
         */
        synchronized (this) {
            notifyAll();

            /*
             * if we're paused, wait for wakeup call
             */
            if (state == STATE_PAUSED) {
                try {
                    wait();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    /*
     * user control stuff
     */

    public void enqueueGameAction(int gameAction) {
        if (state == STATE_RUNNING) {
            synchronized (actionLock) {
                actionCode = gameAction;
                actionLock.notifyAll();
            }
        }
    }

    /*
     * game logic
     */

    private void commitDescent() {
        // TODO: this probably needs synchronisation with processKeyQueue()
        tileCurrent.moveDown();

        tileCurrent.update();
        canvas.flushGraphics();
    }

    public void run() {
        while (stateRequested != STATE_DEAD) {

            /*
             * game action
             */

            if (stateRequested != state)
                setState(stateRequested);

            /*
             * wait for user input
             */
            synchronized (actionLock) {
                try {
                    actionLock.wait(300);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }

                if (actionCode != 0)
                    processAction();
            }
        }

        setState(STATE_DEAD);
    }

    private void processAction() {
        // we're synchronized in here, so size cannot change
        switch (actionCode) {
            case Canvas.UP:
                tileCurrent.rotate();
                break;

            case Canvas.DOWN:
                tileCurrent.moveDown();
                break;

            case Canvas.LEFT:
                tileCurrent.moveLeft();
                break;

            case Canvas.RIGHT:
                tileCurrent.moveRight();
                break;

            case Canvas.FIRE:
                tileCurrent.drop();
                break;
        }
        actionCode = 0;
        tileCurrent.update();
        canvas.flushGraphics();
    }

    public void jumpLevel() {
        // TODO: this should be synchronized
        timerGravity.cancel();

        ++level;
        levelDelay = (levelDelay << 3) / 10 ;

        canvas.renderLevel();

        // TODO: this should probably be done in canvas
        canvas.clearBar();

        timerGravity = new Timer();
        timerGravity.scheduleAtFixedRate(new TimerTask() {

            public void run() {
                commitDescent();
            }

        }, 0, levelDelay);
    }

    public void tileLanded(Tile tile, int pointsAwarded) {

        // check for level jump
        final int tilesInLevel = ++tileCount % TILES_PER_LEVEL;
        if (tilesInLevel == 0)
            jumpLevel();

        canvas.renderBar(tilesInLevel);

        if (pointsAwarded > 10) {
            // redraw field if row was removed
            // TODO: only redraw changed region
            canvas.field.render();
        } else {
            // one last update to be safe
            tileCurrent.update();
        }

        // update tiles
        tileCurrent.clone(tileNext);

        tileNext.randomize();

        tileCurrent.startDescent();

        canvas.field.clearPreview();
        canvas.field.renderTilePreview(tileNext);

        // award points
        score += pointsAwarded;

        // render new score
        canvas.renderScore();
    }

    public void gameOver() {
        stateRequested = STATE_DEAD;
    }

    /*
     * Public Interface
     */

    public void kill() {
        requestState(STATE_DEAD);
    }

    public void resume() {
        requestState(STATE_RUNNING);
    }

    public void pause() {
        requestState(STATE_PAUSED);
    }

    /*
     * serialization 
     */

    public void serialize(DataOutputStream dos) throws IOException {
      tileCurrent.serialize(dos);
      tileNext.serialize(dos);
      dos.writeInt(level);
      dos.writeInt(levelDelay);
      dos.writeInt(score);
      dos.writeInt(tileCount);
    }

    public static GameThread deserialize(UbertrisCanvas canvas, DataInputStream dis) throws IOException {
      return new GameThread(canvas, Tile.deserialize(canvas, dis), Tile.deserialize(canvas, dis),
        dis.readInt(), dis.readInt(), dis.readInt(), dis.readInt());
    }
}
