/**
 * $Id: Tile.java 33 2006-03-22 18:22:13Z Gerhard $
 * 
 * Copyright (C) 2006 Gerhard Zlabinger (gerhard.zlabinger@gmail.com)
 *
 * This file is part of Ubertris.
 * 
 * Ubertris is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ubertris is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ubertris; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

package ubertris;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import java.util.Random;

final class Tile {

    /*
     * static stuff
     */

    private static final int             ROT_0           = 0;
    private static final int             ROT_90          = 1;
    private static final int             ROT_180         = 2;
    private static final int             ROT_270         = 3;

    private static final boolean[][][][] LAYOUTS = new boolean[][][][] {
      { { { true, true, true, true } },
      
        { { true },
          { true },
          { true },
          { true } },
          
        { { true, true, true, true } },
        
        { { true },
          { true },
          { true },
          { true } } },
          
      { { { true, true },
          { true, true } },
          
        { { true, true },
          { true, true } },
    
        { { true, true },
          { true, true } },
          
        { { true, true },
          { true, true } } },
          
      { { { true, true, false },
          { false, true, true } },
          
        { { false, true },
          { true, true },
          { true, false } },
          
        { { true, true, false },
          { false, true, true } },
          
        { { false, true },
          { true, true },
          { true, false } } },
    
      { { { false, true, true },
          { true, true, false } },
          
        { { true, false },
          { true, true },
          
          { false, true } },
        { { false, true, true },
          { true, true, false } },
          
        { { true, false },
          { true, true },
          { false, true } } },
          
      { { { true, true, true },
          { false, false, true } },
            
        { { false, true },
          { false, true },
          { true, true } },
            
        { { true, false, false },
          { true, true, true } },
          
        { { true, true },
          { true, false },
          { true, false } } },
          
      { { { false, false, true },
          { true, true, true } },
          
        { { true, false },
          { true, false },
          { true, true } },
          
        { { true, true, true },
          { true, false, false } },
        { { true, true },
          { false, true },
          { false, true } } },
          
      { { { false, true, false },
          { true, true, true } },
          
        { { true, false },
          { true, true },
          { true, false } },
          
        { { true, true, true },
          { false, true, false } },
          
        { { false, true },
          { true, true },
          { false, true } } } };

    private static final int[]           COLORS          = new int[] { 0xEFEF00, 0xEF0000, 0x00EF00, 0x0000EF, 0x3090A0,
            0x907A30, 0x9000EF                          };

    private static Random                rnd             = new Random();

    /*
     * instance stuff
     */

    boolean[][]     layout;
    int             layoutIdx;
    int             rotation;
    int             posX;
    int             posY;
    int             color;

    int             height;
    int             width;
    boolean         hasLanded;

    boolean[][] oldLayout;
    int         oldPosX;
    int         oldPosY;
    int         oldHeight;
    int         oldWidth;

    private final UbertrisCanvas  tetrisCanvas;

    public Tile(UbertrisCanvas canvas, int layout, int rotation) {
        this.layoutIdx = layout;
        this.layout = LAYOUTS[layout][rotation];
        this.rotation = rotation;
        this.color = COLORS[layoutIdx];

        height = this.layout.length;
        width = this.layout[0].length;

        // TODO: performance by making this static
        tetrisCanvas = canvas;
    }

    public static Tile getRandomTile(UbertrisCanvas canvas) {
        return new Tile(canvas, Math.abs(rnd.nextInt() % 7), Math.abs(rnd.nextInt() % 4));
    }

    public void startDescent() {
        oldPosY = posY = -1;
        oldPosX = posX = (tetrisCanvas.field.WIDTH - width) >> 1;
        oldWidth = width;
        oldHeight = height;
        oldLayout = layout;
        hasLanded = false;
    }

    private boolean canMoveTo(boolean[][] layout, int x, int y) {
        final int width = layout[0].length;

        if (x < 0 || x + width > tetrisCanvas.field.WIDTH || y >= tetrisCanvas.field.HEIGHT)
            return false;

        final int height = layout.length;

        for (int v = height - 1, j = y; j >= 0 && v >= 0; --j, --v) {
            for (int u = 0, i = x; u < width && i < tetrisCanvas.field.WIDTH; ++i, ++u) {
                if (layout[v][u] && !tetrisCanvas.field.isCubeFree(i, j))
                    return false;
            }
        }

        return true;
    }

    private boolean canMoveTo(int x, int y) {
        return canMoveTo(layout, x, y);
    }

    /**
     * @param movX
     * @param movY
     * @return true if the tile could be moved
     */
    public boolean move(int movX, int movY) {
        synchronized (this) {
            if (hasLanded || !canMoveTo(posX + movX, posY + movY))
                return false;

            posY += movY;
            posX += movX;

            return true;
        }
    }

    public boolean moveDown() {
        synchronized (this) {
            if (!move(0, 1)) {
                hasLanded = true;

                if (posY < (height - 1)) {
                    tetrisCanvas.gameOver();
                    return false;
                }

                int score = 10;
                int numRows = 0;

                // update field raster
                for (int v = height - 1, j = posY; j >= 0 && v >= 0; --j, --v) {
                    for (int u = 0, i = posX; u < width && i < tetrisCanvas.field.WIDTH; ++i, ++u) {
                        if (layout[v][u]) {
                            if (tetrisCanvas.field.setColorAt(i, j, color)) {
                                ++numRows;
                                ++j;
                                break;
                            }
                        }
                    }
                }

                switch (numRows) {
                    case 1:
                        score += 10;
                        break;

                    case 2:
                        score += 30;
                        break;

                    case 3:
                        score += 60;
                        break;

                    case 4:
                        score += 90;
                        break;
                }

                tetrisCanvas.tileLanded(this, score);

                return false;
            }

            return true;
        }
    }

    public boolean moveLeft() {
        return move(-1, 0);
    }

    public boolean moveRight() {
        return move(1, 0);
    }

    public void drop() {
        synchronized (this) {
            while (moveDown())
                ;
        }
    }

    public boolean rotate() {
        synchronized (this) {
            final int newRotation = rotation - 1 < 0 ? ROT_270 : rotation - 1;
            final boolean[][] newLayout = LAYOUTS[layoutIdx][newRotation];

            int ofsX = 0, ofsY = 0;

            if (width == 4) {
                ofsX = 2;
                ofsY = 2;
            } else if (height == 4) {
                ofsY = -2;
                ofsX = -2;
            }

            if (!canMoveTo(newLayout, posX + ofsX, posY + ofsY)) {
                return false;
            }

            rotation = newRotation;
            layout = newLayout;
            width = layout[0].length;
            height = layout.length;
            posX += ofsX;
            posY += ofsY;

            return true;
        }
    }

    public void clone(Tile from) {
        layoutIdx = from.layoutIdx;
        layout = from.layout;
        width = from.width;
        height = from.height;
        rotation = from.rotation;
        color = from.color;
    }

    public void randomize() {
        layoutIdx = Math.abs(rnd.nextInt() % 7);
        rotation = Math.abs(rnd.nextInt() % 4);
        layout = LAYOUTS[layoutIdx][rotation];
        color = COLORS[layoutIdx];
        width = layout[0].length;
        height = layout.length;
    }

    public synchronized void update() {
        tetrisCanvas.field.eraseTile(this);
        tetrisCanvas.field.renderTile(this);

        oldWidth = width;
        oldHeight = height;
        oldPosX = posX;
        oldPosY = posY;
        oldLayout = layout;
    }

    public boolean hasLanded() {
        return hasLanded;
    }

    /*
     * Serialization
     */

    public void serialize(DataOutputStream dos) throws IOException {
        dos.writeInt(layoutIdx);
        dos.writeInt(rotation);
        dos.writeInt(posX);
        dos.writeInt(posY);
    }

    public static Tile deserialize(UbertrisCanvas canvas, DataInputStream dis) throws IOException {
        final Tile tile = new Tile(canvas, dis.readInt(), dis.readInt());
        
        tile.posX = dis.readInt();
        tile.posY = dis.readInt();

        return tile;
    }
}
